/** gps2sys:
  *
  *  A simple (even simple-minded) utility to set the Windows(R) system time from
  *  the text output of a Garmin(R) GPS unit over a serial cable.
  *
  *  Use at your own risk: I don't guarantee that this will work, or whether or
  *  not it will crash your system. (Well, I'm pretty sure it won't, but no
  *  guarantees!)
  *
  *  Usage:
  *  - Set the GPS up to send Garmin standard text output out the serial line.
  *  - Compile - using "cl /MT gps2sys.cpp" in windows
  *  - Run, supplying serial port as first argument
  *
  *  @version 0.01a
  *  @author Dusty deBoer; ddeboer(at)k-state.edu
  *  @credits to: http://www.codeproject.com/system/serial_com.asp
  **/
#include <stdio.h>
#include <io.h>
#include <windows.h>
#include <process.h>
#include <stddef.h>
#include <stdlib.h>
#include <conio.h>
#include <fcntl.h>

#define MSGMAX 		100		// Maximum length of string from GPS
#define POLLSMAX 	4		// Poll the GPS POLLSMAX times for the current time
#define FUDGEMS  	5		// Add FUDGEMS to GPS time to compensate for serial line latency
#define TIMEOUT		10000	// ms to wait until timing out
#define DEF_BAUD    9600
#define DEF_BYTE    8
#define DEF_PRTY    0
#define DEF_SBTS    0

void timeout(void* dummy) {
	Sleep(TIMEOUT);
	printf("Timeout waiting to read port: check port name and GPS configuration.\n");
	exit(1);
}

int main(int argc, char** argv) {
  SYSTEMTIME syst;
  SYSTEMTIME gpst;
  int polls;
  unsigned long msgend, bytes, line;
  char msg[MSGMAX];
  char *commport = NULL;
  int failed = 1;
  int retval = 0;
  int dcb_baud = DEF_BAUD;
  int dcb_byte = DEF_BYTE;
  int dcb_prty = DEF_PRTY;
  int dcb_sbts = DEF_SBTS;
  if (argc<2) {
	  printf("Usage: %s <serialport> [<baud> <bytesize> <parity> <stopbits>]\n");
	  printf("Example: \"%s COM1 9600 8 0 0\"\n",argv[0],argv[0]);
	  return 1;
  } else {
	  commport = argv[1];
	  if (argc>2) {
		  dcb_baud = atoi(argv[2]);
		  dcb_byte = atoi(argv[3]);
		  dcb_prty = atoi(argv[4]);
		  dcb_sbts = atoi(argv[5]);
	  }
  }
  polls = 0;
  _beginthread(timeout, 0, NULL );
  HANDLE m_hCommPort = ::CreateFile(argv[1],GENERIC_READ,0,0,OPEN_EXISTING,0,0);
  DCB dcb = {0};
  dcb.DCBlength = sizeof(DCB);
  if (!::GetCommState(m_hCommPort,&dcb)) {
	  printf("Failed to get comm state: %d\n",GetLastError());
	  return E_FAIL;
  }
  dcb.BaudRate = dcb_baud;
  dcb.ByteSize = dcb_byte;
  dcb.Parity   = dcb_prty;
  dcb.StopBits = dcb_sbts;
  SetCommState(m_hCommPort, &dcb);
  while (polls<POLLSMAX) {
	  msgend = 0;
	  bytes  = 0;
	  line   = 0;
	  while (msgend<MSGMAX && line==0) {
		  //bytes = _read(portfd,(msg+msgend),(MSGMAX-msgend));
		  if (!ReadFile(m_hCommPort,(msg+msgend),(MSGMAX-msgend),&bytes,NULL)) { bytes=0; }
		  while(bytes>0) {
			  if (msg[msgend]=='\r'||msg[msgend]=='\n') { bytes=0; line=1; }
			  else { bytes--; msgend++; }
		  }
		  msg[msgend]='\0';
	  }
	  GetSystemTime(&syst);
	  if (msgend>0 && sscanf(msg,"@%2d%2d%2d%2d%2d%2d",
		  &gpst.wYear,&gpst.wMonth,&gpst.wDay,&gpst.wHour,&gpst.wMinute,&gpst.wSecond)==6) {
			failed = 0;
			gpst.wYear = gpst.wYear + ((int)(syst.wYear/100))*100;
			// calculate approx time to transmit msgsend bytes over serial connection...
			// this is approximately the time in ms from the GPS time, at this point.
			// plus, add some system processing time (FUDGEMS)
	  		gpst.wMilliseconds = (((float)dcb_byte/dcb_baud)*msgend*1000)+FUDGEMS; // fudge things a bit.
	  } else {
		  failed = 1;
	  }
	  printf(" %s %s\n",(failed?"ERR:":"OK: "),msg);
	  polls++;
  }
  CloseHandle(m_hCommPort);

  printf("\n");
  if (failed) {
	printf("Failed to get time from GPS: check output format.\n");
  } else {
	printf("GPS time:    %04d-%02d-%02d %02d:%02d:%02d %04d UTC\n",
		gpst.wYear,gpst.wMonth,gpst.wDay,gpst.wHour,gpst.wMinute,gpst.wSecond,gpst.wMilliseconds);
	printf("System time: %04d-%02d-%02d %02d:%02d:%02d %04d UTC\n",
		syst.wYear,syst.wMonth,syst.wDay,syst.wHour,syst.wMinute,syst.wSecond,syst.wMilliseconds);
	if((syst.wYear-gpst.wYear)>1||(gpst.wYear-syst.wYear)>1) {
		printf(" Time differs by more than 1 year; aborting.\n");
		retval = 1;
	} else if(SetSystemTime(&gpst)) {
		printf(" Successfully set system time.\n");
		retval = 0;
	} else {
		printf(" Failed to set system time.\n");
		retval = 1;
	}
  }

  return retval;
}

